'''
Simple test for opencl4py, edited from the example distributed in that package.
'''

import os
import opencl4py as cl
import numpy as np
os.environ["PYOPENCL_CTX"] = "0:0" # This is where you set which devices are in the context
# EVERY machine will have a device 0:0, which by default is the CPU
# Other devices will have various numbers
# Substitute her the device you want to compare to the CPU
from time import clock

if __name__ == "__main__":
    
    N = 20000000 # Size of vectors to work with
    use_DP = True # Whether to use double precision floating point
    
    # Print all of the platforms and devices to screen
    platforms = cl.Platforms()
    print('List of all platforms and devices:')
    print(platforms.dump_devices())
    print('')
    
    # Create a context and a queue
    ctx = platforms.create_some_context()
    queue = ctx.create_queue(ctx.devices[0]) # This is where you choose a device
    
    # Tell the user about the test that will be run
    device_can_use_DP = ('cl_khr_fp64' in queue.device.extensions) or ('cl_amd_fp64' in queue.device.extensions)
    print('Will test OpenCL on ' + queue.device.name + '.')
    if use_DP and device_can_use_DP:
        print('The test will use double precision, which this device can handle.')
    elif (not use_DP) and device_can_use_DP:
        print('The test will use single precision, but the device is capable of double precision.')
    elif use_DP and (not device_can_use_DP):
        print ('The test would use double precision, but the device is not capable; expect an error.')
    elif (not use_DP) and (not device_can_use_DP):
        print('The test will use single precision.  This device is not capable of double precision.')
    print('Test vectors have ' + str(N) + ' elements each.\n')
    
    
    # Define a simple kernel using double precision floating point
    double_code = """
        #pragma OPENCL EXTENSION cl_khr_fp64 : enable
        #pragma OPENCL EXTENSION cl_amd_fp64 : enable
        
        __kernel void test(__global const double *a, __global const double *b,
                           __global double *c, const double k) {
          size_t i = get_global_id(0);
          c[i] = (a[i] + b[i]) * k;
        }
        """
        
    # Define a simple kernel using single precision floating point
    single_code = """
        __kernel void test(__global const float *a, __global const float *b,
                           __global float *c, const float k) {
          size_t i = get_global_id(0);
          c[i] = (a[i] + b[i]) * k;
        }
        """
    
    # Define the kernel using single or double precision as selected above
    if use_DP:
        my_type = np.float64
        prg = ctx.create_program(double_code)
    else:
        my_type = np.float32
        prg = ctx.create_program(single_code)
    krn = prg.get_kernel("test")
    
    # Define numpy arrays with arbitrary numbers
    a = np.arange(N, dtype=my_type) # Input array a
    b = np.arange(N, dtype=my_type) # Input array b
    c = np.empty(N, dtype=my_type)  # Input array c (which will hold result of calculation)
    k = np.array([0.8], dtype=my_type) # Constant scalar
    
    # Define OpenCL memory buffers, passing appropriate flags and inputs
    a_buf = ctx.create_buffer(cl.CL_MEM_READ_ONLY | cl.CL_MEM_COPY_HOST_PTR, a) # Read only, copy values from host, use array a
    b_buf = ctx.create_buffer(cl.CL_MEM_READ_ONLY | cl.CL_MEM_COPY_HOST_PTR, b) # Read only, copy values from host, use array b
    c_buf = ctx.create_buffer(cl.CL_MEM_WRITE_ONLY | cl.CL_MEM_ALLOC_HOST_PTR, size=c.nbytes) # Write only, allocate memory, use byte size of array c
    
    # Run the kernel and time it
    t_start = clock()
    krn.set_args(a_buf, b_buf, c_buf, k[0:1]) # Set kernel arguments as the three buffers and a float
    queue.execute_kernel(krn, [N], None) # Execute the simple kernel, specifying the global workspace dimensionality and local workspace dimensionality (None uses some default)  
    queue.read_buffer(c_buf, c) # Read the memory buffer for c into the numpy array for c 
    t_end = clock()
    print('OpenCL took ' + str(t_end-t_start) + ' seconds.')
    
    # Now do the equivalent work as the kernel, but in Python (and time it)
    t_start = clock()
    truth = (a + b) * k[0]
    t_end = clock()
    print('Python took ' + str(t_end-t_start) + ' seconds.')
    
    # Make sure that OpenCL and Python actually agree on their results
    max_diff = np.max(np.fabs(c - truth))
    print("Maximum difference between OpenCL and Python calculations is " +  str(max_diff))



"""
Functions for working with the discrete-continuous EGM (DCEGM) algorithm as
described in "The endogenous grid method for discrete-continuous dynamic
choice models with (or without) taste shocks" by Iskhakov et al. (2016)
[https://doi.org/10.3982/QE643 and ijrsDCEGM2017 in our Zotero]
"""
import numpy as np
from HARK.interpolation import LinearInterp


def calcSegments(x, v):
    """
    Find index vectors `rise` and `fall` such that `rise` holds the indeces `i`
    such that x[i+1]>x[i] and `fall` holds indeces `j` such that either
    - x[j+1] < x[j] or,
    - x[j]>x[j-1] and v[j]<v[j-1].

    The vectors are essential to the DCEGM algorithm, as they definite the
    relevant intervals to be used to construct the upper envelope of potential
    solutions to the (necessary) first order conditions.

    Parameters
    ----------
    x : np.ndarray
        array of points where `v` is evaluated
    v : np.ndarray
        array of values of some function of `x`

    Returns
    -------
    rise : np.ndarray
        see description above
    fall : np.ndarray
        see description above
    """
    # NOTE: assumes that the first segment is in fact increasing (forced in EGM
    # by augmentation with the constrained segment).
    # elements in common grid g

    # Identify index intervals of falling and rising regions
    # We need these to construct the upper envelope because we need to discard
    # solutions from the inverted Euler equations that do not represent optimal
    # choices (the FOCs are only necessary in these models).
    #
    # `fall` is a vector of indeces that represent the first elements in all
    # of the falling segments (the curve can potentially fold several times)
    fall = np.empty(0, dtype=int)  # initialize with empty and then add the last point below while-loop

    rise = np.array([0])  # Initialize such thatthe lowest point is the first grid point
    i = 1  # Initialize
    while i <= len(x) - 2:
        # Check if the next (`ip1` stands for i plus 1) grid point is below the
        # current one, such that the line is folding back.
        ip1_falls = x[i+1] < x[i]  # true if grid decreases on index increment
        i_rose = x[i] > x[i-1]  # true if grid decreases on index decrement
        val_fell = v[i] < v[i-1]  # true if value rises on index decrement

        if (ip1_falls and i_rose) or (val_fell and i_rose):

            # we are in a region where the endogenous grid is decreasing or
            # the value function rises by stepping back in the grid.
            fall = np.append(fall, i)  # add the index to the vector

            # We now iterate from the current index onwards until we find point
            # where resources rises again. Unfortunately, we need to check
            # each points, as there can be multiple spells of falling endogenous
            # grids, so we cannot use bisection or some other fast algorithm.
            k = i
            while x[k+1] < x[k]:
                k = k + 1
            # k now holds either the next index the starts a new rising
            # region, or it holds the length of M, `m_len`.

            rise = np.append(rise, k)

            # Set the index to the point where resources again is rising
            i = k

        i = i + 1

    # Add the last index for convenience (then all segments are complete, as
    # len(fall) == len(rise), and we can form them by range(rise[j], fall[j]+1).
    fall = np.append(fall, len(v)-1)

    return rise, fall
# think! nanargmax makes everythign super ugly because numpy changed the wraning
# in all nan slices to a valueerror...it's nans, aaarghgghg


def calcMultilineEnvelope(M, C, V_T, commonM):
    """
    Do the envelope step of the DCEGM algorithm. Takes in market ressources,
    consumption levels, and inverse values from the EGM step. These represent
    (m, c) pairs that solve the necessary first order conditions. This function
    calculates the optimal (m, c, v_t) pairs on the commonM grid.

    Parameters
    ----------
    M : np.array
        market ressources from EGM step
    C : np.array
        consumption from EGM step
    V_T : np.array
        transformed values at the EGM grid
    commonM : np.array
        common grid to do upper envelope calculations on

    Returns
    -------


    """
    m_len = len(commonM)
    rise, fall = calcSegments(M, V_T)

    num_kinks = len(fall)  # number of kinks / falling EGM grids

    # Use these segments to sequentially find upper envelopes. commonVARNAME
    # means the VARNAME evaluated on the common grid with a cloumn for each kink
    # discovered in calcSegments. This means that commonVARNAME is a matrix
    # common grid length-by-number of segments to consider. In the end, we'll
    # use nanargmax over the columns to pick out the best (transformed) values.
    # This is why we fill the arrays with np.nan's.
    commonV_T = np.empty((m_len, num_kinks))
    commonV_T[:] = np.nan
    commonC = np.empty((m_len, num_kinks))
    commonC[:] = np.nan

    # Now, loop over all segments as defined by the "kinks" or the combination
    # of "rise" and "fall" indeces. These (rise[j], fall[j]) pairs define regions
    for j in range(num_kinks):
        # Find points in the common grid that are in the range of the points in
        # the interval defined by (rise[j], fall[j]).
        below = M[rise[j]] >= commonM  # boolean array of bad indeces below
        above = M[fall[j]] <= commonM  # boolen array of bad indeces above
        in_range = below + above == 0  # pick out elements that are neither

        # create range of indeces in the input arrays
        idxs = range(rise[j], fall[j]+1)
        # grab ressource values at the relevant indeces
        m_idx_j = M[idxs]

        # based in in_range, find the relevant ressource values to interpolate
        m_eval = commonM[in_range]

        # re-interpolate to common grid
        commonV_T[in_range, j] = LinearInterp(m_idx_j, V_T[idxs], lower_extrap=True)(m_eval) # NOQA
        commonC[in_range, j]   = LinearInterp(m_idx_j, C[idxs], lower_extrap=True)(m_eval) # NOQA Interpolat econsumption also. May not be nesserary
    # for each row in the commonV_T matrix, see if all entries are np.nan. This
    # would mean that we have no valid value here, so we want to use this boolean
    # vector to filter out irrelevant entries of commonV_T.
    row_all_nan = np.array([np.all(np.isnan(row)) for row in commonV_T])
    # Now take the max of all these line segments.
    idx_max = np.zeros(commonM.size, dtype=int)
    idx_max[row_all_nan == False] = np.nanargmax(commonV_T[row_all_nan == False], axis=1)

    # prefix with upper for variable that are "upper enveloped"
    upperV_T = np.zeros(m_len)

    # Set the non-nan rows to the maximum over columns
    upperV_T[row_all_nan == False] = np.nanmax(commonV_T[row_all_nan == False, :], axis=1)
    # Set the rest to nan
    upperV_T[row_all_nan] = np.nan

    # Add the zero point in the bottom
    if np.isnan(upperV_T[0]):
        # in transformed space space, utility of zero-consumption (-inf) is 0.0
        upperV_T[0] = 0.0
        # commonM[0] is typically 0, so this is safe, but maybe it should be 0.0
        commonC[0] = commonM[0]

    # Extrapolate if NaNs are introduced due to the common grid
    # going outside all the sub-line segments
    IsNaN = np.isnan(upperV_T)
    upperV_T[IsNaN] = LinearInterp(commonM[IsNaN == False], upperV_T[IsNaN == False])(commonM[IsNaN])
    LastBeforeNaN = np.append(np.diff(IsNaN) > 0, 0)
    LastId = LastBeforeNaN*idx_max  # Find last id-number
    idx_max[IsNaN] = LastId[IsNaN]
    # Linear index used to get optimal consumption based on "id"  from max
    ncols = commonC.shape[1]
    rowidx = np.cumsum(ncols*np.ones(len(commonM), dtype=int))-ncols
    idx_linear = np.unravel_index(rowidx+idx_max, commonC.shape)
    upperC = commonC[idx_linear]
    upperC[IsNaN] = LinearInterp(commonM[IsNaN == 0], upperC[IsNaN == 0])(commonM[IsNaN])

    # TODO calculate cross points of line segments to get the true vertical drops

    upperM = commonM.copy()  # anticipate this TODO

    return upperM, upperC, upperV_T


def main():
    print("Sorry, HARK.dcegm doesn't actually do anything on its own.")


if __name__ == '__main__':
    main()



"""
This file tests whether HARK's models are initialized correctly.
"""


# Bring in modules we need
import unittest
import numpy as np
import HARK.ConsumptionSaving.ConsumerParameters as Params
from HARK.ConsumptionSaving.ConsIndShockModel import PerfForesightConsumerType
from HARK.ConsumptionSaving.ConsIndShockModel import KinkedRconsumerType
from HARK.ConsumptionSaving.ConsIndShockModel import IndShockConsumerType
from HARK.ConsumptionSaving.ConsMarkovModel import MarkovConsumerType
from HARK.utilities import plotFuncsDer, plotFuncs
from copy import copy

class testInitialization(unittest.TestCase):
    # We don't need a setUp method for the tests to run, but it's convenient
    # if we want to test various things on the same model in different test_*
    # methods.
    def test_PerfForesightConsumerType(self):
        try:
            model = PerfForesightConsumerType(**Params.init_perfect_foresight)
        except:
            self.fail("PerfForesightConsumerType failed to initialize with Params.init_perfect_foresight.")

    def test_IndShockConsumerType(self):
        try:
            model = IndShockConsumerType(**Params.init_lifecycle)
        except:
            self.fail("IndShockConsumerType failed to initialize with Params.init_lifecycle.")

    def test_KinkedRconsumerType(self):
        try:
            model = KinkedRconsumerType(**Params.init_kinked_R)
        except:
            self.fail("KinkedRconsumerType failed to initialize with Params.init_kinked_R.")

    def test_MarkovConsumerType(self):
        try:
            unemp_length = 5 # Averange length of unemployment spell
            urate_good = 0.05        # Unemployment rate when economy is in good state
            urate_bad = 0.12         # Unemployment rate when economy is in bad state
            bust_prob = 0.01         # Probability of economy switching from good to bad
            recession_length = 20    # Averange length of bad state
            p_reemploy =1.0/unemp_length
            p_unemploy_good = p_reemploy*urate_good/(1-urate_good)
            p_unemploy_bad = p_reemploy*urate_bad/(1-urate_bad)
            boom_prob = 1.0/recession_length
            MrkvArray = np.array([[(1-p_unemploy_good)*(1-bust_prob),p_unemploy_good*(1-bust_prob),
                                   (1-p_unemploy_good)*bust_prob,p_unemploy_good*bust_prob],
                                  [p_reemploy*(1-bust_prob),(1-p_reemploy)*(1-bust_prob),
                                   p_reemploy*bust_prob,(1-p_reemploy)*bust_prob],
                                  [(1-p_unemploy_bad)*boom_prob,p_unemploy_bad*boom_prob,
                                   (1-p_unemploy_bad)*(1-boom_prob),p_unemploy_bad*(1-boom_prob)],
                                  [p_reemploy*boom_prob,(1-p_reemploy)*boom_prob,
                                   p_reemploy*(1-boom_prob),(1-p_reemploy)*(1-boom_prob)]])

            # Make a consumer with serially correlated unemployment, subject to boom and bust cycles
            init_serial_unemployment = copy(Params.init_idiosyncratic_shocks)
            init_serial_unemployment['MrkvArray'] = [MrkvArray]
            init_serial_unemployment['UnempPrb'] = 0 # to make income distribution when employed
            init_serial_unemployment['global_markov'] = False
            SerialUnemploymentExample = MarkovConsumerType(**init_serial_unemployment)
        except:
            self.fail("MarkovConsumerType failed to initialize with boom/bust unemployment.")



'''
This file appears to be an old version of what is now ConsRepAgentModel.py.
Its previous contents have been entirely removed and replaced with a universal
import from ConsRepAgentModel.  Whenever a user imports from this file, they
will get a warning that they should import from ConsRepAgentModel instead.
'''

import warnings
from HARK.ConsumptionSaving.ConsRepAgentModel import *

warnings.warn('Please import from ConsRepAgentModel rather than RepAgentModel.  This module will be removed in a future version of HARK.')


import unittest

from HARK.validators import non_empty

class ValidatorsTests(unittest.TestCase):
    '''
    Tests for validator decorators which validate function arguments
    '''

    def test_non_empty(self):
        @non_empty('list_a')
        def foo(list_a, list_b):
            pass

        try:
            foo([1], [])
        except Exception:
            self.fail()
        with self.assertRaisesRegexp(
            TypeError,
            'Expected non-empty argument for parameter list_a',
        ):
            foo([], [1])

        @non_empty('list_a', 'list_b')
        def foo(list_a, list_b):
            pass

        with self.assertRaisesRegexp(
            TypeError,
            'Expected non-empty argument for parameter list_b',
        ):
            foo([1], [])
        with self.assertRaisesRegexp(
            TypeError,
            'Expected non-empty argument for parameter list_a',
        ):
            foo([], [1])



'''
Decorators which can be used for validating arguments passed into decorated functions
'''

from __future__ import print_function

import sys
from functools import wraps

if sys.version_info[0] < 3:
    from funcsigs import signature
else:
    from inspect import signature


def non_empty(*parameter_names):
    '''
    Enforces arguments to parameters passed in have len > 0
    '''

    def _decorator(f):
        sig = signature(f)
        # TODO - add validation that parameter names are in signature

        @wraps(f)
        def _inner(*args, **kwargs):
            bindings = sig.bind(*args, **kwargs)
            for parameter_name in parameter_names:
                if not len(bindings.arguments[parameter_name]):
                    raise TypeError(
                        'Expected non-empty argument for parameter {}'.format(parameter_name)
                    )
            return f(*args, **kwargs)
        return _inner
    return _decorator
